<?php
/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */

class CBbotmaker extends ChatBroker {

    public $url             = 'https://api.botmaker.com/v2.0';
    public $field_chatid    = 'sessionId';
    public $field_from      = 'fromName';
    public $field_body      = 'message';
    public $field_number    = 'contactId';
    public $field_to        = 'to';
    public $ACCESS_TOKEN    = '';
    public $CLIENT_ID       = '';
    public $SECRET_ID       = '';
    public $PUBLIC_URL      = '';
    public $DID             = '';
    public $ICON            = '../menu/images/whatsapp.webp';
    public $config_template = array( "CLIENT_ID"=> "", "SECRET_ID" => "", "ACCESS_TOKEN" => "", "PUBLIC_URL"=>"" );
    public $allow_originate = true;
    public $use_number_from = 'whatsapp';
    public $field_help      = array(
      'did'=>'This field should be the full whatsapp number including country code, but without the plus sign.',
      'template'=>'CLIENT_ID and SECRET_ID are supplied in the Botmaker platform. ACCESS_TOKEN can be retrieved also from the BotMaker platform. PUBLIC_URL is the public URL of your Chatbroker instance.',
      'engine'=>'Whatsapp API from https://botmaker.com'
    );

    public function get_hook_url() {
        $did      = $this->get('DID');
        $clientid = $this->get('CLIENT_ID');
        $secretid = $this->get('SECRET_ID');
        $urldata = array('CLIENT_ID'=>$clientid,'DID'=>$did);
        $webhookurl = $_SERVER['REQUEST_SCHEME']."://".$_SERVER['HTTP_HOST'].dirname($_SERVER['SCRIPT_NAME'])."/hook.php?botmaker=1&to={DID}";
        $webhookurl = $this->replace_variables($webhookurl,$urldata);
        return $webhookurl;
    }

    public function filter_data($data) {
        unset($data['from']); // from is always "user", we want to use fromName instead
        if(isset($data['image'])) {
            $data['attach']=$this->download_attach($data,'image');
        }
        return $data;
    }

    public function download_attach($data,$type) {
        $url = $data[$type];
        $parsedUrl = parse_url($url);
        $filename = basename($parsedUrl['path']);
 
        $fp = fopen("/tmp/".$filename,'w+');
        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_FILE, $fp);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_exec($ch);
        $statusCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);
        fclose($fp);

        if($statusCode == 200) {
            $mime = mime_content_type("/tmp/$filename");
            if($mime=='application/ogg') {
                $mime='audio/ogg';
            }
            return array("mime"=>$mime,"file"=>"/tmp/$filename","filename"=>$filename);
        }
        return array();
    }

    /*

    public function set_web_hook() {
    }


    public function preprocess_data($data) {
        return $data;
    }

    public function postprocess_data($data,$result) {
        // inject new values into data array if needed
        $return = array();
        return $return;
    }

    public function getAvatar($phone,$data=array()) {
        return $basepath."/avatars/$phone.$ext";
    }
    */

    public function checkSendOk($server_output) {
        $result = json_decode($server_output,1);
        if($result['success']=='true') {
            return true;
        } else {
            return false;
        }
    }

    public function send($from,$fromdid,$to,$body,$msgqueueid,$provider,$conversationid,$context) {

        $this->cdebug("botmaker send from $from, fromdid $fromdid, to $to, body $body, msgqueueid $msgqueueid");

        $purl       = $this->get('PUBLIC_URL');
        $channel    = substr(get_class($this),2);
        $token      = $this->get('ACCESS_TOKEN');
        $url        = $this->url;

        if($msgqueueid<>'') {
            $files = $this->check_for_attachments($msgqueueid);
        } else {
            $files = array();
        }

/*
{
  "chat": {
    "channelId": "botproject-whatsapp-5491147038***",
    "contactId": "5491147038***"
  },
  "messages": [
    {
      "text": "string",
      "buttons": [
        {
          "label": "string",
          "intentIdOrName": "string",
          "webhookPayload": "string",
          "description": "string"
        }
      ],
      "media": {
        "mimeType": "image/jpeg",
        "url": "string"
      },
      "location": {
        "latitude": "string",
        "longitude": "string",
        "name": "string",
        "address": "string"
      },
      "contact": {
        "firstName": "string",
        "middleName": "string",
        "lastName": "string",
        "emails": [
          "string"
        ],
        "phones": [
          "string"
        ]
      },
      "webhookPayload": "string"
    }
  ]
}
*/

        if(count($files)==0) {
            if(!is_array($body)) {
                $url   .= "/chats-actions/send-messages";
                $this->cdebug("botmaker url $url, no attachments, body $body");
                $ch = curl_init();
                curl_setopt($ch, CURLOPT_URL, $url);
                curl_setopt($ch, CURLOPT_HTTPHEADER, array('access-token: '.$token,'Content-Type:application/json'));
                curl_setopt($ch, CURLOPT_POST, 1);
                $myjson = json_encode(array('chat' => array("channelId"=>$fromdid ,"contactId"=>$to), "messages" => array(array("text"=>$body))));
                curl_setopt($ch, CURLOPT_POSTFIELDS, $myjson);
                curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
                $server_output = curl_exec($ch);
                curl_close($ch);
                //file_put_contents("/tmp/borrame.log","token $token\n",FILE_APPEND);
                //file_put_contents("/tmp/borrame.log","$myjson\n",FILE_APPEND);

            } else {
                // TODO button messages
                $body_text = $body['body']['text'];
                $body_text.="\n\n";
                $this->cdebug("botmaker url $url, reply buttons, body $body_text");
                foreach($body['action']['buttons'] as $idx=>$val) {
                    $body_text.=$val['reply']['title'];
                    $body_text.="\n";
                }
                $this->cdebug("botmaker url $url, reply buttons, body $body_text");
                $ch = curl_init();
                curl_setopt($ch, CURLOPT_URL, $url);
                curl_setopt($ch, CURLOPT_HTTPHEADER, array('authorization: Bearer '.$token,'Content-Type:application/json'));
                curl_setopt($ch, CURLOPT_POST, 1);
                curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode(array('chatId' => $to.'@c.us', 'message' => $body_text)));
                curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
                $server_output = curl_exec($ch);
                $this->cdebug("botmaker output after post: $server_output");
                curl_close($ch);
            }
        } else {
            $url   .= "/chats-actions/send-messages";
            foreach($files as $file) {

                // retrieve file to temp disk to get metadata (mime type, image info, etc)
                $path = $this->chatbroker_host ."/api/files/messages/".$msgqueueid."/".$file;
                $type = pathinfo($path, PATHINFO_EXTENSION);
                $data = file_get_contents($path);
                file_put_contents("/tmp/$file",$data);
                $image_info = getimagesize("/tmp/$file");
                $mime = mime_content_type("/tmp/$file");
                if($mime=='application/ogg') {
                    $mime='audio/ogg';
                }
                unlink("/tmp/$file");

                $purl .= "/realtime/api/files/messages/".$msgqueueid."/".$file;

                $this->cdebug("botmaker url $url, no attachments, body $body");
                $ch = curl_init();
                curl_setopt($ch, CURLOPT_URL, $url);
                curl_setopt($ch, CURLOPT_HTTPHEADER, array('access-token: '.$token,'Content-Type:application/json'));
                curl_setopt($ch, CURLOPT_POST, 1);
                $myjson = json_encode(array('chat' => array("channelId"=>$fromdid ,"contactId"=>$to), "messages" => array(array("text"=>$body,"media"=>array("mimeType"=>"$mime","url"=>"$purl")))));
                curl_setopt($ch, CURLOPT_POSTFIELDS, $myjson);
                curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
                $server_output = curl_exec($ch);
                curl_close($ch);
                //file_put_contents("/tmp/borrame.log","token $token\n",FILE_APPEND);
                //file_put_contents("/tmp/borrame.log","$myjson\n",FILE_APPEND);
            }
        }

        $this->cdebug("botmaker server output follows: ".rtrim($server_output));
        $chatid = $to; // in botmaker, the chatid is the destination number in outgoing
        $this->mdr('outbound',$channel,$from,$fromdid,$to,$conversationid,$body,$type,$provider,$context);

        $send_result = json_decode($server_output,1);

        if(isset($send_result['requestId'])) {
            return '{"success":"true"}';
        } else {
            return '{"success":"false"}';
        }
    }

  public function filter_number($number) {
      $pertes = preg_split("/@/",$number);
      $partes = preg_split("/\./",$pertes[0]);
      return $partes[0];
  }

  public function filter_chatid($chatid) {
      $pertes = preg_split("/@/",$chatid);
      $partes = preg_split("/\./",$pertes[0]);
      return "botmaker_".$partes[0];
  }

  public function process_request($req,$payload) {
      $data = array_merge($req,json_decode($payload,1));
      return $data;
  }

}
